@@ -9,7 +9,8 @@
   Copyright 2016      Sjjad Hashemian
   Copyright 2016      Stephen Thirlwall
   Copyright 2017      Antonio Quartulli
-  Copyright 2017      Ismael Luceno
+  Copyright 2017-2018 Ismael Luceno
+  Copyright 2018      Shankar
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
@@ -38,11 +39,15 @@
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-*/
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ */
 
 /* Connection stuff */
 
+#include "config.h"
 #include "axel.h"
+#include "hash.h"
 
 /**
  * Convert an URL to a conn_t structure.
@@ -92,8 +97,6 @@ conn_set(conn_t *conn, const char *set_url)
 	} else {
 		*i = 0;
 		snprintf(conn->dir, MAX_STRING, "/%s", i + 1);
-		if (conn->proto == PROTO_HTTP || conn->proto == PROTO_HTTPS)
-			http_encode(conn->dir, sizeof(conn->dir));
 	}
 	j = strchr(conn->dir, '?');
 	if (j != NULL)
@@ -181,7 +184,7 @@ scheme_from_proto(int proto)
 }
 
 /* Generate a nice URL string. */
-char *
+int
 conn_url(char *dst, size_t len, conn_t *conn)
 {
 	const char *prefix = "", *postfix = "";
@@ -190,7 +193,7 @@ conn_url(char *dst, size_t len, conn_t *conn)
 
 	size_t scheme_len = strlcpy(dst, scheme, len);
 	if (scheme_len > len)
-		return NULL;
+		return -1;
 
 	len -= scheme_len;
 
@@ -199,7 +202,7 @@ conn_url(char *dst, size_t len, conn_t *conn)
 	if (*conn->user != 0 && strcmp(conn->user, "anonymous") != 0) {
 		int plen = snprintf(p, len, "%s:%s@", conn->user, conn->pass);
 		if (plen < 0)
-			return NULL;
+			return -1;
 		len -= plen;
 		p += plen;
 	}
@@ -209,11 +212,8 @@ conn_url(char *dst, size_t len, conn_t *conn)
 		postfix = "]";
 	}
 
-	int plen;
-	plen = snprintf(p, len, "%s%s%s:%i%s%s", prefix, conn->host, postfix,
-			conn->port, conn->dir, conn->file);
-
-	return plen < 0 ? NULL : dst;
+	return snprintf(p, len, "%s%s%s:%i%s%s", prefix, conn->host, postfix,
+	                conn->port, conn->dir, conn->file);
 }
 
 /* Simple... */
@@ -271,16 +271,21 @@ conn_init(conn_t *conn)
 		if (!http_connect(conn->http, conn->proto, proxy, conn->host,
 				  conn->port, conn->user, conn->pass,
 				  conn->conf->io_timeout)) {
-			conn->message = conn->http->headers;
+			conn->message = conn->http->headers->p;
 			conn_disconnect(conn);
 			return 0;
 		}
-		conn->message = conn->http->headers;
+		conn->message = conn->http->headers->p;
 		conn->tcp = &conn->http->tcp;
 	}
 	return 1;
 }
 
+/**
+ * Setup the connection.
+ *
+ * Must be called with the conn_t lock held.
+ */
 int
 conn_setup(conn_t *conn)
 {
@@ -289,13 +294,13 @@ conn_setup(conn_t *conn)
 			return 0;
 
 	if (PROTO_IS_FTP(conn->proto) && !conn->proxy) {
-		/* Set up data connnection */
+		/* Set up data connection */
 		if (!ftp_data(conn->ftp, conn->conf->io_timeout))
 			return 0;
 		conn->tcp = &conn->ftp->data_tcp;
 
 		if (conn->currentbyte) {
-			ftp_command(conn->ftp, "REST %lld", conn->currentbyte);
+			ftp_command(conn->ftp, "REST %jd", conn->currentbyte);
 			if (ftp_wait(conn->ftp) / 100 != 3 &&
 			    conn->ftp->status / 100 != 2)
 				return 0;
@@ -308,6 +313,8 @@ conn_setup(conn_t *conn)
 		conn->http->firstbyte =
 			conn->supported ? conn->currentbyte : -1;
 		conn->http->lastbyte = conn->lastbyte;
+
+		abuf_setup(conn->http->request, 2048);
 		http_get(conn->http, s);
 		for (i = 0; i < conn->conf->add_header_count; i++)
 			http_addheader(conn->http, "%s",
@@ -324,123 +331,210 @@ conn_exec(conn_t *conn)
 			return 0;
 		return ftp_wait(conn->ftp) / 100 == 1;
 	} else {
+		abuf_setup(conn->http->headers, 1024);
 		if (!http_exec(conn->http))
 			return 0;
 		return conn->http->status / 100 == 2;
 	}
 }
 
+static
+int
+conn_info_ftp(conn_t *conn)
+{
+	ftp_command(conn->ftp, "REST %d", 1);
+	if (ftp_wait(conn->ftp) / 100 == 3 ||
+	    conn->ftp->status / 100 == 2) {
+		conn->supported = true;
+		ftp_command(conn->ftp, "REST %d", 0);
+		ftp_wait(conn->ftp);
+	} else {
+		conn->supported = false;
+	}
+
+	if (!ftp_cwd(conn->ftp, conn->dir))
+		return 0;
+	conn->size = ftp_size(conn->ftp, conn->file,
+			      conn->conf->max_redirect,
+			      conn->conf->io_timeout);
+	if (conn->size < 0)
+		conn->supported = false;
+	if (conn->size == -1)
+		return 0;
+	else if (conn->size == -2)
+		conn->size = LLONG_MAX;
+
+	return 1;
+}
+
+struct urlseq {
+	uint64_t secret[1];
+	char buf[MAX_STRING];
+	int num_urls;
+	uint32_t visited_urls[];
+};
+
+static
+void *
+urlseq_teardown(struct urlseq **u)
+{
+	free(*u);
+	return *u = NULL;
+}
+
+static
+struct urlseq *
+urlseq_init(unsigned int urlcount)
+{
+	struct urlseq *u;
+	u = malloc(sizeof(*u) + sizeof(u->visited_urls[0]) * urlcount);
+	if (!u || axel_rand64(u->secret) < (ssize_t)sizeof(u->secret[0]))
+		return urlseq_teardown(&u);
+	u->num_urls = 0;
+	return u;
+}
+
+static
+int
+urlseq_check_loop(struct urlseq *u, conn_t *conn)
+{
+	if (!u)
+		return 0;
+
+	int url_len = conn_url(u->buf, sizeof(u->buf), conn);
+	if (url_len <= 0)
+		return 0;
+
+	uint32_t url_hash = axel_hash32(u->buf, url_len, u->secret);
+	for (int j = 0; j < u->num_urls; j++) {
+		if (u->visited_urls[j] == url_hash)
+			return 1;
+	}
+	u->visited_urls[u->num_urls++] = url_hash;
+	return 0;
+}
+
 /* Get file size and other information */
 int
 conn_info(conn_t *conn)
 {
 	/* It's all a bit messed up.. But it works. */
 	if (PROTO_IS_FTP(conn->proto) && !conn->proxy) {
-		ftp_command(conn->ftp, "REST %d", 1);
-		if (ftp_wait(conn->ftp) / 100 == 3 ||
-		    conn->ftp->status / 100 == 2) {
-			conn->supported = true;
-			ftp_command(conn->ftp, "REST %d", 0);
-			ftp_wait(conn->ftp);
-		} else {
-			conn->supported = false;
+		return conn_info_ftp(conn);
+	}
+
+	char s[1005];
+	long long int i = 0;
+
+	struct urlseq *urlseq = urlseq_init(conn->conf->max_redirect);
+
+	do {
+		const char *t;
+
+		conn->supported = true;
+		conn->currentbyte = 0;
+		pthread_mutex_lock(&conn->lock);
+		int setup_ret = conn_setup(conn);
+		pthread_mutex_unlock(&conn->lock);
+		if (!setup_ret)
+			return 0;
+		conn_exec(conn);
+		conn_disconnect(conn);
+
+		http_filename(conn->http, conn->output_filename);
+
+		/* Code 3xx == redirect */
+		if (conn->http->status / 100 != 3)
+			break;
+		if ((t = http_header(conn->http, "location:")) == NULL)
+			return 0;
+		sscanf(t, "%1000s", s);
+		if (s[0] == '/') {
+			abuf_printf(conn->http->headers, "%s%s:%i%s",
+				    scheme_from_proto(conn->proto),
+				    conn->host, conn->port, s);
+			strlcpy(s, conn->http->headers->p, sizeof(s));
+		} else if (strstr(s, "://") == NULL) {
+			conn_url(conn->http->headers->p,
+				 conn->http->headers->len,
+				 conn);
+			strlcat(conn->http->headers->p, s,
+				conn->http->headers->len);
+			strlcpy(s, conn->http->headers->p, sizeof(s));
 		}
 
-		if (!ftp_cwd(conn->ftp, conn->dir))
+		if (!conn_set(conn, s)) {
 			return 0;
-		conn->size = ftp_size(conn->ftp, conn->file,
-				      conn->conf->max_redirect,
-				      conn->conf->io_timeout);
-		if (conn->size < 0)
-			conn->supported = false;
-		if (conn->size == -1)
-			return 0;
-		else if (conn->size == -2)
-			conn->size = LLONG_MAX;
-	} else {
-		char s[1005];
-		long long int i = 0;
-
-		do {
-			const char *t;
-
-			conn->supported = true;
-			conn->currentbyte = 0;
-			if (!conn_setup(conn))
-				return 0;
-			conn_exec(conn);
-			conn_disconnect(conn);
-
-			http_filename(conn->http, conn->output_filename);
-
-			/* Code 3xx == redirect */
-			if (conn->http->status / 100 != 3)
-				break;
-			if ((t = http_header(conn->http, "location:")) == NULL)
-				return 0;
-			sscanf(t, "%1000s", s);
-			if (s[0] == '/') {
-				snprintf(conn->http->headers,
-					 sizeof(conn->http->headers),
-					 "%s%s:%i%s",
-					 scheme_from_proto(conn->proto),
-					 conn->host, conn->port, s);
-				strlcpy(s, conn->http->headers, sizeof(s));
-			} else if (strstr(s, "://") == NULL) {
-				conn_url(conn->http->headers,
-					 sizeof(conn->http->headers), conn);
-				strlcat(conn->http->headers, s,
-					sizeof(conn->http->headers));
-				strlcpy(s, conn->http->headers, sizeof(s));
-			}
-
-			if (!conn_set(conn, s)) {
-				return 0;
-			}
-
-			/* check if the download has been redirected to FTP and
-			 * report it back to the caller */
-			if (PROTO_IS_FTP(conn->proto) && !conn->proxy) {
-				return -1;
-			}
-
-			i++;
 		}
-		while (conn->http->status / 100 == 3 &&
-		       i < conn->conf->max_redirect);
 
-		if (i == conn->conf->max_redirect) {
+		/* check if the download has been redirected to FTP and
+		 * report it back to the caller */
+		if (PROTO_IS_FTP(conn->proto) && !conn->proxy) {
+			return -1;
+		}
+
+		if (++i >= conn->conf->max_redirect) {
 			fprintf(stderr, _("Too many redirects.\n"));
 			return 0;
 		}
 
-		conn->size = http_size_from_range(conn->http);
-		/* We assume partial requests are supported if a Content-Range
-		 * header is present.
-		 */
-		conn->supported = conn->http->status == 206 || conn->size > 0;
+		/* Check if the current URL has already been visited */
+		if (urlseq_check_loop(urlseq, conn)) {
+			fprintf(stderr, _("Redirect loop detected.\n"));
+			return 0;
+		}
+	} while (conn->http->status / 100 == 3);
 
-		/* If Content-Length and Content-Range disagree, it's a server
-		 * bug; we take the larger and hope for the best.
-		 */
-		conn->size = max(conn->size, http_size(conn->http));
+	/* Free the memory allocated for the redirect loop detection */
+	urlseq_teardown(&urlseq);
 
-		if (conn->size <= 0) {
-			/* Sanity check */
-			switch (conn->http->status) {
-			case 200:
-			case 206:
-			case 416:
-				break;
-			default: /* unexpected */
-				return 0;
-			}
+	/* Check for non-recoverable errors */
+	if (conn->http->status != 416 && conn->http->status / 100 != 2)
+		return 0;
 
-			/* So we got an invalid or no size, fall back */
+	conn->size = http_size_from_range(conn->http);
+	/* We assume partial requests are supported if a Content-Range
+	 * header is present.
+	 */
+	conn->supported = conn->http->status == 206 || conn->size > 0;
+
+	if (conn->size <= 0) {
+		/* Sanity check */
+		switch (conn->http->status) {
+		case 200: /* OK -> unsupported */
+		case 416: /* Range Not Satisfiable -> broken? */
 			conn->supported = false;
-			conn->size = LLONG_MAX;
+		case 206: /* Partial Content */
+			break;
+		default: /* unexpected */
+			return 0;
 		}
 	}
-
+	/* If Content-Range is missing or disagrees with Content-Length, it's a
+	 * server bug; we take the larger and hope for the best.
+	 */
+	conn->size = max(conn->size, http_size(conn->http));
+	if (conn->size <= 0) {
+		conn->size = LLONG_MAX;
+		conn->supported = false;
+	}
 	return 1;
 }
+
+int
+conn_info_status_get(char *msg, size_t size, conn_t *conn)
+{
+	if (is_proto_http(conn->proto)) {
+		char *p = conn->http->headers->p;
+		/* Skip protocol and code */
+		while (*p++ != ' ');
+		while (*p++ != ' ');
+		size_t len = strcspn(p, "\r\n");
+		if (len) {
+			strlcpy(msg, p, min(len + 1, size));
+			return conn->http->status;
+		}
+	}
+	strlcpy(msg, _("Unknown Error"), size);
+	return -1;
+}
