@@ -35,7 +35,9 @@
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-*/
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ */
 
 /* Connection stuff */
 
@@ -70,6 +72,13 @@
 #define PROTO_DEFAULT          PROTO_HTTP
 #define PROTO_DEFAULT_PORT     PROTO_HTTP_PORT
 
+static inline
+int
+is_proto_http(int proto)
+{
+	return (proto & PROTO_PROTO_MASK) == PROTO_PROTO_HTTP;
+}
+
 typedef struct {
 	conf_t *conf;
 
@@ -85,9 +94,9 @@ typedef struct {
 
 	ftp_t ftp[1];
 	http_t http[1];
-	long long int size;	/* File size, not 'connection size'.. */
-	long long int currentbyte;
-	long long int lastbyte;
+	off_t size; /* File size, not 'connection size'.. */
+	off_t currentbyte;
+	off_t lastbyte;
 	tcp_t *tcp;
 	bool enabled;
 	bool supported;
@@ -101,12 +110,13 @@ typedef struct {
 } conn_t;
 
 int conn_set(conn_t *conn, const char *set_url);
-char *conn_url(char *dst, size_t len, conn_t *conn);
+int conn_url(char *dst, size_t len, conn_t *conn);
 void conn_disconnect(conn_t *conn);
 int conn_init(conn_t *conn);
 int conn_setup(conn_t *conn);
 int conn_exec(conn_t *conn);
 int conn_info(conn_t *conn);
+int conn_info_status_get(char *msg, size_t size, conn_t *conn);
 const char *scheme_from_proto(int proto);
 
 #endif				/* AXEL_CONN_H */
