@@ -11,7 +11,8 @@
   Copyright 2016      Stephen Thirlwall
   Copyright 2017      Antonio Quartulli
   Copyright 2017      David Polverari
-  Copyright 2017      Ismael Luceno
+  Copyright 2017-2019 Ismael Luceno
+  Copyright 2018-2019 Shankar
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
@@ -40,12 +41,17 @@
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-*/
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ */
 
 /* HTTP control file */
 
+#include "config.h"
 #include "axel.h"
 
+#define HDR_CHUNK 512
+
 inline static int
 is_default_port(int proto, int port)
 {
@@ -72,8 +78,10 @@ http_auth_token(char *token, const char *user, const char *pass)
 	const char *auth[] = { user, ":", pass, NULL };
 	const char **p = auth;
 
-	while (*p && **p) {
+	while (*p) {
 		char a = chain_next(&p);
+		if (!a)
+			break;
 		*token++ = base64_encode[a >> 2];
 		char b = chain_next(&p);
 		*token++ = base64_encode[((a & 3) << 4) | (b >> 4)];
@@ -105,24 +113,17 @@ http_connect(http_t *conn, int proto, char *proxy, char *host, int port,
 	conn->port = port;
 	conn->proto = proto;
 
-	if (proxy != NULL) {
-		if (*proxy != 0) {
-			snprintf(conn->host, sizeof(conn->host),
-				 "%s:%i", host, port);
-			if (!conn_set(tconn, proxy)) {
-				fprintf(stderr,
-					_("Invalid proxy string: %s\n"), proxy);
-				return 0;
-			}
-			host = tconn->host;
-			port = tconn->port;
-			proto = tconn->proto;
-			puser = tconn->user;
-			ppass = tconn->pass;
-			conn->proxy = 1;
-		} else {
-			conn->proxy = 0;
+	if (proxy && *proxy) {
+		if (!conn_set(tconn, proxy)) {
+			fprintf(stderr, _("Invalid proxy string: %s\n"), proxy);
+			return 0;
 		}
+		host = tconn->host;
+		port = tconn->port;
+		proto = tconn->proto;
+		puser = tconn->user;
+		ppass = tconn->pass;
+		conn->proxy = 1;
 	}
 
 	if (tcp_connect(&conn->tcp, host, port, PROTO_IS_SECURE(proto),
@@ -161,7 +162,7 @@ http_get(http_t *conn, char *lurl)
 		postfix = "]";
 	}
 
-	*conn->request = 0;
+	*conn->request->p = 0;
 	if (conn->proxy) {
 		const char *proto = scheme_from_proto(conn->proto);
 		if (is_default_port(conn->proto, conn->port)) {
@@ -188,13 +189,13 @@ http_get(http_t *conn, char *lurl)
 		http_addheader(conn, "Proxy-Authorization: Basic %s",
 			       conn->proxy_auth);
 	http_addheader(conn, "Accept: */*");
-	if (conn->firstbyte >= 0) {
-		if (conn->lastbyte)
-			http_addheader(conn, "Range: bytes=%lld-%lld",
-				       conn->firstbyte, conn->lastbyte);
-		else
-			http_addheader(conn, "Range: bytes=%lld-",
-				       conn->firstbyte);
+	http_addheader(conn, "Accept-Encoding: identity");
+	if (conn->lastbyte && conn->firstbyte >= 0) {
+		http_addheader(conn, "Range: bytes=%jd-%jd",
+			       conn->firstbyte, conn->lastbyte - 1);
+	} else if (conn->firstbyte >= 0) {
+		http_addheader(conn, "Range: bytes=%jd-",
+			       conn->firstbyte);
 	}
 }
 
@@ -209,42 +210,44 @@ http_addheader(http_t *conn, const char *format, ...)
 	strlcat(s, "\r\n", sizeof(s));
 	va_end(params);
 
-	strlcat(conn->request, s, sizeof(conn->request));
+	if (abuf_strcat(conn->request, s) < 0) {
+		fprintf(stderr, "Out of memory\n");
+	}
 }
 
 int
 http_exec(http_t *conn)
 {
-	int i = 0;
-	ssize_t nwrite = 0;
-	char s[2] = {0}, *s2;
+	char *s2;
 
-#ifdef DEBUG
+#ifndef NDEBUG
 	fprintf(stderr, "--- Sending request ---\n%s--- End of request ---\n",
-		conn->request);
+		conn->request->p);
 #endif
 
-	strlcat(conn->request, "\r\n", sizeof(conn->request));
+	strlcat(conn->request->p, "\r\n", conn->request->len);
 
-	while (nwrite < (ssize_t)strlen(conn->request)) {
-		if ((i =
-		     tcp_write(&conn->tcp, conn->request + nwrite,
-			       strlen(conn->request) - nwrite)) < 0) {
+	const size_t reqlen = strlen(conn->request->p);
+	size_t nwrite = 0;
+	while (nwrite < reqlen) {
+		ssize_t tmp;
+		tmp = tcp_write(&conn->tcp, conn->request->p + nwrite,
+				reqlen - nwrite);
+		if (tmp < 0) {
 			if (errno == EINTR || errno == EAGAIN)
 				continue;
-
 			fprintf(stderr,
 				_("Connection gone while writing.\n"));
 			return 0;
 		}
-		nwrite += i;
+		nwrite += tmp;
 	}
 
-	*conn->headers = 0;
+	*conn->headers->p = 0;
 
 	/* Read the headers byte by byte to make sure we don't touch the
 	   actual data */
-	while (1) {
+	for (char *s = conn->headers->p;;) {
 		if (tcp_read(&conn->tcp, s, 1) <= 0) {
 			fprintf(stderr, _("Connection gone.\n"));
 			return 0;
@@ -253,25 +256,41 @@ http_exec(http_t *conn)
 		if (*s == '\r') {
 			continue;
 		} else if (*s == '\n') {
-			if (i == 0)
+			if (s > conn->headers->p && s[-1] == '\n') {
+				*s = 0;
 				break;
-			i = 0;
-		} else {
-			i++;
+			}
+		}
+		s++;
+
+		size_t pos = s - conn->headers->p;
+		if (pos + 10 < conn->headers->len) {
+			int tmp = abuf_setup(conn->headers,
+					     conn->headers->len + HDR_CHUNK);
+			if (tmp < 0) {
+				fprintf(stderr, "Out of memory\n");
+				return 0;
+			}
+			s = conn->headers->p + pos;
 		}
-		/* FIXME wasteful */
-		strlcat(conn->headers, s, sizeof(conn->headers));
 	}
 
-#ifdef DEBUG
+#ifndef NDEBUG
 	fprintf(stderr, "--- Reply headers ---\n%s--- End of headers ---\n",
-		conn->headers);
+		conn->headers->p);
 #endif
 
-	sscanf(conn->headers, "%*s %3i", &conn->status);
-	s2 = strchr(conn->headers, '\n');
-	*s2 = 0;
-	strlcpy(conn->request, conn->headers, sizeof(conn->request));
+	sscanf(conn->headers->p, "%*s %3i", &conn->status);
+	s2 = strchr(conn->headers->p, '\n');
+	if (s2)
+		*s2 = 0;
+	const size_t reslen = s2 - conn->headers->p + 1;
+	if (conn->request->len < reqlen) {
+		int ret = abuf_setup(conn->request, reslen);
+		if (ret < 0)
+			return 0;
+	}
+	memcpy(conn->request->p, conn->headers->p, reslen);
 	*s2 = '\n';
 
 	return 1;
@@ -280,7 +299,7 @@ http_exec(http_t *conn)
 const char *
 http_header(const http_t *conn, const char *header)
 {
-	const char *p = conn->headers;
+	const char *p = conn->headers->p;
 	size_t hlen = strlen(header);
 
 	do {
@@ -296,54 +315,68 @@ http_header(const http_t *conn, const char *header)
 	return NULL;
 }
 
-long long int
+off_t
 http_size(http_t *conn)
 {
 	const char *i;
-	long long int j;
+	off_t j;
 
 	if ((i = http_header(conn, "Content-Length:")) == NULL)
 		return -2;
 
-	sscanf(i, "%lld", &j);
+	sscanf(i, "%jd", &j);
 	return j;
 }
 
-long long int
+off_t
 http_size_from_range(http_t *conn)
 {
 	const char *i;
-	long long int j;
-
 	if ((i = http_header(conn, "Content-Range:")) == NULL)
 		return -2;
 
 	i = strchr(i, '/');
-	if (i == NULL)
+	if (!i++)
 		return -2;
 
-	if (sscanf(i + 1, "%lld", &j) != 1)
+	off_t j = strtoll(i, NULL, 10);
+	if (!j && *i != '0')
 		return -3;
 
 	return j;
 }
 
+/**
+ * Extract file name from Content-Disposition HTTP header.
+ *
+ * Header format:
+ * Content-Disposition: inline
+ * Content-Disposition: attachment
+ * Content-Disposition: attachment; filename="filename.jpg"
+ */
 void
 http_filename(const http_t *conn, char *filename)
 {
 	const char *h;
 	if ((h = http_header(conn, "Content-Disposition:")) != NULL) {
-		sscanf(h, "%*s%*[ \t]filename%*[ \t=\"\'-]%254[^\n\"\' ]",
+		sscanf(h, "%*s%*[ \t]filename%*[ \t=\"\'-]%254[^\n\"\']",
 		       filename);
+		/* Trim spaces at the end of string */
+		const char space[] = "\t ";
+		for (char *n, *p = filename; (p = strpbrk(p, space)); p = n) {
+			n = p + strspn(p, space);
+			if (!*n) {
+				*p = 0;
+				break;
+			}
+		}
 
 		/* Replace common invalid characters in filename
 		   https://en.wikipedia.org/wiki/Filename#Reserved_characters_and_words */
-		char *i = filename;
-		const char *invalid_characters = "/\\?%*:|<>";
+		const char invalid[] = "/\\?%*:|<>";
 		const char replacement = '_';
-		while ((i = strpbrk(i, invalid_characters)) != NULL) {
+		for (char *i = filename; (i = strpbrk(i, invalid)); i++) {
 			*i = replacement;
-			i++;
 		}
 	}
 }
@@ -391,26 +424,3 @@ http_decode(char *s)
 	} while (*s == '%');
 	*p = 0;
 }
-
-void
-http_encode(char *s, size_t len)
-{
-	char t[MAX_STRING];
-	unsigned i, j;
-
-	for (i = j = 0; s[i] && j < sizeof(t) - 1; i++, j++) {
-		t[j] = s[i];
-		if (s[i] <= 0x20 || s[i] >= 0x7f) {
-			/* Fix buffer overflow */
-			if (j >= sizeof(t) - 3) {
-				break;
-			}
-
-			encode_byte(t + j, s[i]);
-			j += 2;
-		}
-	}
-	t[j] = 0;
-
-	strlcpy(s, t, len);
-}
