@@ -4,7 +4,7 @@
   Copyright 2016      Sjjad Hashemian
   Copyright 2016-2017 Stephen Thirlwall
   Copyright 2017      Antonio Quartulli
-  Copyright 2017      Ismael Luceno
+  Copyright 2017-2019 Ismael Luceno
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
@@ -33,15 +33,25 @@
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-*/
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ */
 
 /* SSL interface */
 
-#include "axel.h"
-
-#ifdef HAVE_SSL
+#include "config.h"
 
+#ifdef HAVE_WOLFSSL
+#include <wolfssl/options.h>
+#include <wolfssl/wolfcrypt/settings.h>
+#include <wolfssl/openssl/ssl.h>
+#include <wolfssl/openssl/err.h>
+#else
+#include <openssl/ssl.h>
 #include <openssl/err.h>
+#endif
+
+#include "axel.h"
 
 static pthread_mutex_t ssl_lock;
 static bool ssl_inited = false;
@@ -71,7 +81,7 @@ ssl_startup(void)
 SSL *
 ssl_connect(int fd, char *hostname)
 {
-
+	X509 *server_cert;
 	SSL_CTX *ssl_ctx;
 	SSL *ssl;
 
@@ -92,9 +102,37 @@ ssl_connect(int fd, char *hostname)
 	if (err <= 0) {
 		fprintf(stderr, _("SSL error: %s\n"),
 			ERR_reason_error_string(ERR_get_error()));
+		SSL_CTX_free(ssl_ctx);
 		return NULL;
 	}
 
+	if (conf->insecure) {
+		return ssl;
+	}
+
+	err = SSL_get_verify_result(ssl);
+	if (err != X509_V_OK) {
+		fprintf(stderr, _("SSL error: Certificate error\n"));
+		SSL_CTX_free(ssl_ctx);
+		return NULL;
+	}
+
+	server_cert =  SSL_get_peer_certificate(ssl);
+	if (server_cert == NULL) {
+		fprintf(stderr, _("SSL error: Certificate not found\n"));
+		SSL_CTX_free(ssl_ctx);
+		return NULL;
+	}
+
+	if (!ssl_validate_hostname(hostname, server_cert)) {
+		fprintf(stderr, _("SSL error: Hostname verification failed\n"));
+		X509_free(server_cert);
+		SSL_CTX_free(ssl_ctx);
+		return NULL;
+	}
+
+	X509_free(server_cert);
+
 	return ssl;
 }
 
@@ -104,5 +142,3 @@ ssl_disconnect(SSL *ssl)
 	SSL_shutdown(ssl);
 	SSL_free(ssl);
 }
-
-#endif				/* HAVE_SSL */
