@@ -9,7 +9,10 @@
   Copyright 2016      Mridul Malpotra
   Copyright 2016      Stephen Thirlwall
   Copyright 2017      Antonio Quartulli
-  Copyright 2017      Ismael Luceno
+  Copyright 2017-2019 Ismael Luceno
+  Copyright 2019      Evangelos Foutras
+  Copyright 2019      Kun Ma
+
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
@@ -38,23 +41,25 @@
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-*/
+ *
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ */
 
 /* Text interface */
 
-#include "axel.h"
-
+#include "config.h"
 #include <sys/ioctl.h>
+#include "axel.h"
 
 
 static void stop(int signal);
-static char *size_human(char *dst, size_t len, size_t value);
 static char *time_human(char *dst, size_t len, unsigned int value);
-static void print_commas(long long int bytes_done);
+static void print_commas(off_t bytes_done);
 static void print_alternate_output(axel_t *axel);
-static void print_progress(size_t cur, size_t prev, size_t total, double kbps);
+static void print_progress(off_t cur, off_t prev, off_t total, double kbps);
 static void print_help(void);
 static void print_version(void);
+static void print_version_info(void);
 static int get_term_width(void);
 
 int run = 1;
@@ -79,6 +84,7 @@ static struct option axel_options[] = {
 	{"help",            0,      NULL, 'h'},
 	{"version",         0,      NULL, 'V'},
 	{"alternate",       0,      NULL, 'a'},
+	{"percentage",      0,      NULL, 'p'},
 	{"insecure",        0,      NULL, 'k'},
 	{"no-clobber",      0,      NULL, 'c'},
 	{"header",          1,      NULL, 'H'},
@@ -88,6 +94,16 @@ static struct option axel_options[] = {
 };
 #endif
 
+/**
+ * Unified percentage calculation for all progress indicators.
+ */
+static
+unsigned
+calc_percentage(off_t cur, off_t total)
+{
+	return min(100, (100 * cur + total / 2) / total);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -96,7 +112,7 @@ main(int argc, char *argv[])
 	search_t *search;
 	conf_t conf[1];
 	axel_t *axel;
-	int j, cur_head = 0, ret = 1;
+	int j, ret = 1;
 	char *s;
 
 	fn[0] = 0;
@@ -107,6 +123,8 @@ main(int argc, char *argv[])
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 #endif
+	if (axel_rnd_init() == -1)
+		return 1;
 
 	if (!conf_init(conf)) {
 		return 1;
@@ -117,7 +135,7 @@ main(int argc, char *argv[])
 	j = -1;
 	while (1) {
 		int option = getopt_long(argc, argv,
-					 "s:n:o:S::46NqvhVakcH:U:T:",
+					 "s:n:o:S::46NqvhVapkcH:U:T:",
 					 axel_options, NULL);
 		if (option == -1)
 			break;
@@ -128,11 +146,16 @@ main(int argc, char *argv[])
 				      "User-Agent", optarg);
 			break;
 		case 'H':
-			strlcpy(conf->add_header[cur_head++], optarg,
+			if(!(conf->add_header_count<MAX_ADD_HEADERS)) {
+				fprintf(stderr,
+					_("Too many custom headers (-H)! Currently only %u custom headers can be appended.\n"), MAX_ADD_HEADERS-HDR_count_init);
+				goto free_conf;
+			}
+			strlcpy(conf->add_header[conf->add_header_count++], optarg,
 				sizeof(conf->add_header[0]));
 			break;
 		case 's':
-			if (!sscanf(optarg, "%i", &conf->max_speed)) {
+			if (!sscanf(optarg, "%llu", &conf->max_speed)) {
 				print_help();
 				goto free_conf;
 			}
@@ -168,7 +191,10 @@ main(int argc, char *argv[])
 			conf->ai_family = AF_INET;
 			break;
 		case 'a':
-			conf->alternate_output = 1;
+			conf->progress_style = AXEL_PROGRESS_STYLE_ALTERNATIVE;
+			break;
+		case 'p':
+			conf->progress_style = AXEL_PROGRESS_STYLE_PERCENTAGE;
 			break;
 		case 'k':
 			conf->insecure = 1;
@@ -190,7 +216,7 @@ main(int argc, char *argv[])
 				j++;
 			break;
 		case 'V':
-			print_version();
+			print_version_info();
 			ret = 0;
 			goto free_conf;
 		case 'q':
@@ -210,12 +236,11 @@ main(int argc, char *argv[])
 			goto free_conf;
 		}
 	}
-	conf->add_header_count = cur_head;
 
-	/* disable alternate output and verbosity when quiet is specified */
-	if (conf->verbose < 0)
-		conf->alternate_output = 0;
-	else if (j > -1)
+	/* disable alternate outputs and verbosity when quiet is specified */
+	if (conf->verbose < 0) {
+		conf->progress_style = AXEL_PROGRESS_STYLE_CLASSIC;
+	} else if (j > -1)
 		conf->verbose = j;
 
 	if (conf->num_connections < 1) {
@@ -255,7 +280,9 @@ main(int argc, char *argv[])
 		}
 	}
 
-	printf(_("Initializing download: %s\n"), s);
+	if (conf->progress_style != AXEL_PROGRESS_STYLE_PERCENTAGE)
+		printf(_("Initializing download: %s\n"), s);
+
 	if (do_search) {
 		search = calloc(conf->search_amount + 1, sizeof(search_t));
 		if (!search)
@@ -284,7 +311,7 @@ main(int argc, char *argv[])
 			j = min(j, conf->search_top);
 			printf("%-60s %15s\n", "URL", _("Speed"));
 			for (i = 0; i < j; i++)
-				printf("%-70.70s %5i\n", search[i].url,
+				printf("%-70.70s %5jd\n", search[i].url,
 				       search[i].speed);
 			printf("\n");
 		}
@@ -294,20 +321,16 @@ main(int argc, char *argv[])
 			print_messages(axel);
 			goto close_axel;
 		}
-	} else if (argc - optind == 1) {
-		axel = axel_new(conf, 0, s);
-		if (!axel || axel->ready == -1) {
-			print_messages(axel);
-			goto close_axel;
-		}
 	} else {
 		search = calloc(argc - optind, sizeof(search_t));
 		if (!search)
 			goto free_conf;
 
-		for (int i = 0; i < argc - optind; i++)
+		for (int i = 0; i < argc - optind; i++) {
 			strlcpy(search[i].url, argv[optind + i],
 				sizeof(search[i].url));
+			// FIXME check url here
+		}
 		axel = axel_new(conf, argc - optind, search);
 		free(search);
 		if (!axel || axel->ready == -1) {
@@ -379,14 +402,14 @@ main(int argc, char *argv[])
 	axel_start(axel);
 	print_messages(axel);
 
-	if (conf->alternate_output) {
+	if (conf->progress_style == AXEL_PROGRESS_STYLE_ALTERNATIVE
+	    || conf->progress_style == AXEL_PROGRESS_STYLE_PERCENTAGE) {
 		putchar('\n');
-	} else {
-		if (axel->bytes_done > 0) {	/* Print first dots if resuming */
-			putchar('\n');
-			print_commas(axel->bytes_done);
-			fflush(stdout);
-		}
+	} else if (axel->bytes_done > 0) {	/* Print first dots if resuming */
+		putchar('\n');
+		print_commas(axel->bytes_done);
+		fflush(stdout);
+
 	}
 	axel->start_byte = axel->bytes_done;
 
@@ -395,12 +418,15 @@ main(int argc, char *argv[])
 	signal(SIGTERM, stop);
 
 	while (!axel->ready && run) {
-		size_t prev;
+		off_t prev;
 
 		prev = axel->bytes_done;
 		axel_do(axel);
 
-		if (conf->alternate_output) {
+		if (conf->progress_style == AXEL_PROGRESS_STYLE_PERCENTAGE) {
+			if (!axel->message && prev != axel->bytes_done)
+				printf("%u\n", calc_percentage(axel->bytes_done, axel->size));
+		} else 	if (conf->progress_style == AXEL_PROGRESS_STYLE_ALTERNATIVE) {
 			if (!axel->message && prev != axel->bytes_done)
 				print_alternate_output(axel);
 		} else if (conf->verbose > -1) {
@@ -409,18 +435,15 @@ main(int argc, char *argv[])
 		}
 
 		if (axel->message) {
-			if (conf->alternate_output == 1) {
+			if (conf->progress_style == AXEL_PROGRESS_STYLE_ALTERNATIVE) {
 				/* clreol-simulation */
-				putchar('\r');
-				for (int i = get_term_width(); i > 0; i--)
-					putchar(' ');
-				putchar('\r');
+				fputs("\e[2K\r", stdout);
 			} else {
 				putchar('\n');
 			}
 			print_messages(axel);
 			if (!axel->ready) {
-				if (conf->alternate_output != 1)
+				if (conf->progress_style != AXEL_PROGRESS_STYLE_ALTERNATIVE)
 					print_commas(axel->bytes_done);
 				else
 					print_alternate_output(axel);
@@ -433,7 +456,7 @@ main(int argc, char *argv[])
 
 	char hsize[MAX_STRING / 2], htime[MAX_STRING / 2];
 	time_human(htime, sizeof(htime), axel_gettime() - axel->start_time);
-	size_human(hsize, sizeof(hsize), axel->bytes_done - axel->start_byte);
+	axel_size_human(hsize, sizeof(hsize), axel->bytes_done - axel->start_byte);
 
 	printf(_("\nDownloaded %s in %s. (%.2f KB/s)\n"), hsize, htime,
 	       (double)axel->bytes_per_second / 1024);
@@ -449,7 +472,7 @@ main(int argc, char *argv[])
 }
 
 /* SIGINT/SIGTERM handler */
-void
+RETSIGTYPE
 stop(int signal)
 {
 	(void)signal;
@@ -468,7 +491,7 @@ log2i(unsigned long long x)
 
 /* Convert a number of bytes to a human-readable form */
 char *
-size_human(char *dst, size_t len, size_t value)
+axel_size_human(char *dst, size_t len, size_t value)
 {
 	double fval = (double)value;
 	const char * const oname[] = {
@@ -506,7 +529,7 @@ time_human(char *dst, size_t len, unsigned int value)
 /* Part of the infamous wget-like interface. Just put it in a function
 	because I need it quite often.. */
 void
-print_commas(long long int bytes_done)
+print_commas(off_t bytes_done)
 {
 	int i, j;
 
@@ -527,13 +550,13 @@ print_commas(long long int bytes_done)
  */
 static
 void
-print_progress(size_t cur, size_t prev, size_t total, double kbps)
+print_progress(off_t cur, off_t prev, off_t total, double kbps)
 {
 	prev /= 1024;
 	cur /= 1024;
 
 	bool print_speed = prev > 0;
-	for (size_t i = prev; i < cur; i++) {
+	for (off_t i = prev; i < cur; i++) {
 		if (i % 50 == 0) {
 			if (print_speed)
 				printf("  [%6.1fKB/s]", kbps);
@@ -541,8 +564,8 @@ print_progress(size_t cur, size_t prev, size_t total, double kbps)
 			if (total == LLONG_MAX)
 				printf("\n[ N/A]  ");
 			else
-				printf("\n[%3zu%%]  ",
-				       min(100U, 102400 * i / total));
+				printf("\n[%3u%%]  ",
+				       calc_percentage(1024 * i, total));
 		} else if (i % 10 == 0) {
 			putchar(' ');
 		}
@@ -550,10 +573,21 @@ print_progress(size_t cur, size_t prev, size_t total, double kbps)
 	}
 }
 
+static
+char
+alt_id(int n)
+{
+	const char *p = "09AZaz";
+	while (*p && n > p[1] - p[0]) {
+		n -= p[1] - p[0] + 1;
+		p += 2;
+	}
+	return *p ? *p + n : '*';
+}
 
 static void
 print_alternate_output_progress(axel_t *axel, char *progress, int width,
-				long long int done, long long int total,
+				off_t done, off_t total,
 				double now)
 {
 	if (!width)
@@ -566,8 +600,7 @@ print_alternate_output_progress(axel_t *axel, char *progress, int width,
 		if (axel->conn[i].currentbyte < axel->conn[i].lastbyte) {
 			if (now <= axel->conn[i].last_transfer
 				   + axel->conf->connection_timeout / 2) {
-				progress[offset] = i
-					+ (i < 10 ? '0' : ('A' - 10));
+				progress[offset] = alt_id(i);
 			} else
 				progress[offset] = '#';
 		}
@@ -576,15 +609,14 @@ print_alternate_output_progress(axel_t *axel, char *progress, int width,
 	}
 
 	progress[width] = '\0';
-	printf("\r[%3ld%%] [%s", min(100, (long)(done * 100. / total + .5)),
-	       progress);
+	printf("\r[%3u%%] [%s", calc_percentage(done, total), progress);
 }
 
 static void
 print_alternate_output(axel_t *axel)
 {
-	long long int done = axel->bytes_done;
-	long long int total = axel->size;
+	off_t done = axel->bytes_done;
+	off_t total = axel->size;
 	double now = axel_gettime();
 	int width = get_term_width();
 	char *progress;
@@ -592,7 +624,7 @@ print_alternate_output(axel_t *axel)
 	if (width < 40) {
 		fprintf(stderr,
 			_("Can't setup alternate output. Deactivating.\n"));
-		axel->conf->alternate_output = 0;
+		axel->conf->progress_style = AXEL_PROGRESS_STYLE_CLASSIC;
 
 		return;
 	}
@@ -652,6 +684,7 @@ get_term_width(void)
 void
 print_help(void)
 {
+	print_version();
 #ifdef NOGETOPTLONG
 	printf(_("Usage: axel [options] url1 [url2] [url...]\n"
 		 "\n"
@@ -669,6 +702,7 @@ print_help(void)
 		 "-q\tLeave stdout alone\n"
 		 "-v\tMore status information\n"
 		 "-a\tAlternate progress indicator\n"
+		 "-p\tPrint simple percentages instead of progress bar (0-100)\n"
 		 "-h\tThis information\n"
 		 "-T x\tSet I/O and connection timeout\n"
 		 "-V\tVersion information\n"
@@ -692,6 +726,7 @@ print_help(void)
 		 "--quiet\t\t\t-q\tLeave stdout alone\n"
 		 "--verbose\t\t-v\tMore status information\n"
 		 "--alternate\t\t-a\tAlternate progress indicator\n"
+		 "--percentage\t\t-p\tPrint simple percentages instead of progress bar (0-100)\n"
 		 "--help\t\t\t-h\tThis information\n"
 		 "--timeout=x\t\t-T x\tSet I/O and connection timeout\n"
 		 "--version\t\t-V\tVersion information\n"
@@ -703,14 +738,20 @@ print_help(void)
 void
 print_version(void)
 {
-	printf(_("Axel version %s (%s)\n"), VERSION, ARCH);
+	printf(_("Axel %s (%s)\n"), VERSION, ARCH);
+}
+
+void
+print_version_info(void)
+{
+	print_version();
 	printf("\nCopyright 2001-2007 Wilmer van der Gaast,\n"
 	       "\t  2007-2009 Giridhar Appaji Nag,\n"
 	       "\t  2008-2010 Philipp Hagemeister,\n"
 	       "\t  2015-2017 Joao Eriberto Mota Filho,\n"
 	       "\t  2016-2017 Stephen Thirlwall,\n"
-	       "\t  2017      Ismael Luceno,\n"
 	       "\t  2017      Antonio Quartulli,\n"
+	       "\t  2017-2024 Ismael Luceno,\n"
 	       "\t\t    %s\n%s\n\n", _("and others."),
 	       _("Please, see the CREDITS file.\n\n"));
 }
